# ===================================================================
#    ####   #    #  ######   ####     ##
#   #    #  ##  ##  #       #    #   #  #
#   #    #  # ## #  #####   #       #    #
#   #    #  #    #  #       #  ###  ######
#   #    #  #    #  #       #    #  #    #
#    ####   #    #  ######   ####   #    #
#
#
#   #####     ##     ####   ######
#   #    #   #  #   #    #  #
#   #    #  #    #  #       #####
#   #####   ######  #       #
#   #   #   #    #  #    #  #
#   #    #  #    #   ####   ######
#
#    This is a clone of Omega Race for Commodore 64 but with much
#    simpler graphics
#
# ===================================================================
#       Author: Jean-Michel Richer
# Instituation: University of Angers, France
#    Deparment: Computer Science
#        Email: jean-michel.richer@univ-angers.fr
#         Date: February 2023
# ===================================================================

# -------------------------------------------------------------------
# This module defines a moving object
# boxes overlap
# -------------------------------------------------------------------

import math
import copy
import pygame as pg
import geometry as g


"""
    Canvas class that enables to draw objects (see geometry.py)
    on a pygame.Surface.
    The user has to define the (x0,y0) coordonates of the frame
    and a scale factor.

    Each form to draw can be submitted to a rotation and/or a
    translation.

    Each fonction that draws an object returns a new object of 
    the same type with its new coordinates.
"""


class Canvas(object):

    """
    Constructor from pygame.Surface, frame and scale factor
    """

    def __init__(self, window, x0=0, y0=0, scale=1):
        self.window = window
        self.ww = window.get_width()
        self.wh = window.get_height()
        self.x0 = x0
        self.y0 = y0
        self.scale = scale
        self.bounding_box = g.BoundingBox(
            -x0, self.wh - y0, self.ww - x0, -(self.wh - y0)
        )
        print(str(self.bounding_box))

        self.font = pg.font.Font("resources/PermanentMarker-Regular.ttf", 18)

    """
        Return geometry.boundingBox of this Canvas
    """

    def get_bounding_box(self):
        return self.bounding_box

    """
        Draw X and Y axis
    """

    def draw_axis(self, color=(160, 160, 160)):
        pg.draw.line(self.window, color, (0, self.y0), (self.ww, self.y0), 1)
        pg.draw.line(self.window, color, (self.x0, 0), (self.x0, self.wh), 1)

    """
        Draw a bounding box
    """

    def draw_bounding_box(self, bb, color):
        x = self.x0 + min(bb.left, bb.right)
        y = self.y0 - max(bb.top, bb.bottom)
        pg.draw.rect(self.window, color, pg.Rect(x, y, bb.width, bb.height), 1)

    """
        Draw a line with given color and transformations.

        Return a new line with coordinates that have been modified
        if one or several transformations (rotation, translation)
        where performed.

    """

    def draw_line(self, line, color, operations=[]):
        line_copy = copy.deepcopy(line)

        for op in operations:
            if isinstance(op, g.Rotation):
                line_copy.pt1.rotate(op.angle)
                line_copy.pt2.rotate(op.angle)

            if isinstance(op, g.Translation):
                line_copy.pt1.translate(op.pt)
                line_copy.pt2.translate(op.pt)

        line_copy.pt1.x = int(line_copy.pt1.x / self.scale)
        x1 = self.x0 + line_copy.pt1.x
        line_copy.pt2.x = int(line_copy.pt2.x / self.scale)
        x2 = self.x0 + line_copy.pt2.x
        line_copy.pt1.y = int(line_copy.pt1.y / self.scale)
        y1 = self.y0 - line_copy.pt1.y
        line_copy.pt2.y = int(line_copy.pt2.y / self.scale)
        y2 = self.y0 - line_copy.pt2.y

        pg.draw.line(self.window, color, (x1, y1), (x2, y2), 1)

        return line_copy

    """
        Draw a triangle with given color and transformations.

        Return a new triangle with coordinates that have been modified
        if one or several transformations (rotation, translation)
        where performed.
        
    """

    def draw_triangle(self, tri, color, operations=[]):
        triangle_copy = copy.deepcopy(tri)

        for op in operations:
            if isinstance(op, g.Rotation):
                triangle_copy.v1.rotate(op.angle)
                triangle_copy.v2.rotate(op.angle)
                triangle_copy.v3.rotate(op.angle)

            if isinstance(op, g.Translation):
                triangle_copy.v1.translate(op.pt)
                triangle_copy.v2.translate(op.pt)
                triangle_copy.v3.translate(op.pt)

        triangle_copy.v1.x = int(triangle_copy.v1.x / self.scale)
        x1 = self.x0 + triangle_copy.v1.x
        triangle_copy.v2.x = int(triangle_copy.v2.x / self.scale)
        x2 = self.x0 + triangle_copy.v2.x
        triangle_copy.v3.x = int(triangle_copy.v3.x / self.scale)
        x3 = self.x0 + triangle_copy.v3.x

        triangle_copy.v1.y = int(triangle_copy.v1.y / self.scale)
        y1 = self.y0 - triangle_copy.v1.y
        triangle_copy.v2.y = int(triangle_copy.v2.y / self.scale)
        y2 = self.y0 - triangle_copy.v2.y
        triangle_copy.v3.y = int(triangle_copy.v3.y / self.scale)
        y3 = self.y0 - triangle_copy.v3.y

        pg.draw.line(self.window, color, (x1, y1), (x2, y2), 1)
        pg.draw.line(self.window, color, (x2, y2), (x3, y3), 1)
        pg.draw.line(self.window, color, (x3, y3), (x1, y1), 1)

        return triangle_copy

    """
        Draw a rectangle with given color and transformations.

        Return a new rectangle with coordinates that have been modified
        if one or several transformations (rotation, translation)
        where performed.
        
    """

    def draw_rectangle(self, rect, color, operations=[], width=1):
        rectangle_copy = copy.deepcopy(rect)

        for op in operations:
            if isinstance(op, g.Rotation):
                rectangle_copy.tl.rotate(op.angle)
                rectangle_copy.br.rotate(op.angle)
                rectangle_copy.tr.rotate(op.angle)
                rectangle_copy.bl.rotate(op.angle)

            if isinstance(op, g.Translation):
                rectangle_copy.tl.translate(op.pt)
                rectangle_copy.br.translate(op.pt)
                rectangle_copy.tr.translate(op.pt)
                rectangle_copy.bl.translate(op.pt)

        rectangle_copy.tl.x = int(rectangle_copy.tl.x / self.scale)
        x1 = self.x0 + rectangle_copy.tl.x
        rectangle_copy.tr.x = int(rectangle_copy.tr.x / self.scale)
        x2 = self.x0 + rectangle_copy.tr.x
        rectangle_copy.br.x = int(rectangle_copy.br.x / self.scale)
        x3 = self.x0 + rectangle_copy.br.x
        rectangle_copy.bl.x = int(rectangle_copy.bl.x / self.scale)
        x4 = self.x0 + rectangle_copy.bl.x
        rectangle_copy.tl.y = int(rectangle_copy.tl.y / self.scale)
        y1 = self.y0 - rectangle_copy.tl.y
        rectangle_copy.tr.y = int(rectangle_copy.tr.y / self.scale)
        y2 = self.y0 - rectangle_copy.tr.y
        rectangle_copy.br.y = int(rectangle_copy.br.y / self.scale)
        y3 = self.y0 - rectangle_copy.br.y
        rectangle_copy.bl.y = int(rectangle_copy.bl.y / self.scale)
        y4 = self.y0 - rectangle_copy.bl.y

        pg.draw.line(self.window, color, (x1, y1), (x2, y2), width)
        pg.draw.line(self.window, color, (x2, y2), (x3, y3), width)
        pg.draw.line(self.window, color, (x3, y3), (x4, y4), width)
        pg.draw.line(self.window, color, (x4, y4), (x1, y1), width)

        return rectangle_copy

    """
        Draw a circle with given color and transformations.

        Return a new circle with coordinates that have been modified
        if one or several transformations (translation)
        where performed.
        
    """

    def draw_circle(self, circle, color, operations=[]):
        circle_copy = copy.deepcopy(circle)

        for op in operations:
            if isinstance(op, g.Rotation):
                circle_copy.center.rotate(op.angle)

            if isinstance(op, g.Translation):
                circle_copy.center.translate(op.pt)

        circle_copy.center.x = int(circle_copy.center.x / self.scale)
        x1 = self.x0 + circle_copy.center.x
        circle_copy.center.y = int(circle_copy.center.y / self.scale)
        y1 = self.y0 - circle_copy.center.y
        pg.draw.circle(self.window, color, (x1, y1), circle.radius / self.scale, 1)

        return circle_copy

    """
        Draw a polygon with given color and transformations.

        Return a new polygon with coordinates that have been modified
        if one or several transformations (rotation, translation)
        where performed.
        
    """

    def draw_polygon(self, polygon, color, operations=[]):
        polygon_copy = copy.deepcopy(polygon)

        for op in operations:
            if isinstance(op, g.Rotation):
                polygon_copy.rotate(op.angle)

            if isinstance(op, g.Translation):
                polygon_copy.translate(op.pt)

        l = []
        for p in polygon_copy.points:
            p.x = int(p.x / self.scale)
            x = self.x0 + p.x
            p.y = int(p.y / self.scale)
            y = self.y0 - p.y
            l.append((x, y))
        pg.draw.polygon(self.window, color, l, 1)

        return polygon_copy

    def draw_text(self, message, x, y, color=None, font=None):
        if color == None:
            color = (128, 0, 0)

        if font == None:
            font = self.font

        rendu = font.render(message, True, color)
        rectangle = rendu.get_rect()
        x = self.x0 + x
        y = self.y0 - y
        rectangle.topleft = (x, y)
        self.window.blit(rendu, rectangle)
