# ===================================================================
#    ####   #    #  ######   ####     ##
#   #    #  ##  ##  #       #    #   #  #
#   #    #  # ## #  #####   #       #    #
#   #    #  #    #  #       #  ###  ######
#   #    #  #    #  #       #    #  #    #
#    ####   #    #  ######   ####   #    #
#
#
#   #####     ##     ####   ######
#   #    #   #  #   #    #  #
#   #    #  #    #  #       #####
#   #####   ######  #       #
#   #   #   #    #  #    #  #
#   #    #  #    #   ####   ######
#
#    This is a clone of Omega Race for Commodore 64 but with much
#    simpler graphics
#
# ===================================================================
#       Author: Jean-Michel Richer
# Instituation: University of Angers, France
#    Deparment: Computer Science
#        Email: jean-michel.richer@univ-angers.fr
#         Date: February 2023
# ===================================================================

# -------------------------------------------------------------------
# This module defines classes that represent forms :
# - Point, Line, Triangle, Rectangle, Circle and Polygon
# that can be later drawn on a Canvas (see canvas.py)
#
# Those forms can be submitted to transformations such as rotation
# and translation.
# -------------------------------------------------------------------
import math
import copy
import pygame as pg

# constants that define when a contact happens
# as powers of 2
CONTACT_LEFT = 1  # 2^0
CONTACT_RIGHT = 2  # 2^1
CONTACT_TOP = 4  # 2^2
CONTACT_BOTTOM = 8  # 2^3
CONTACT_FULL = 15  # 2^3 + 2^2 + 2^1 + 2^0

where_exactly = [
    0,
    CONTACT_RIGHT + CONTACT_BOTTOM,
    CONTACT_LEFT + CONTACT_BOTTOM,
    CONTACT_BOTTOM,
    CONTACT_RIGHT + CONTACT_TOP,
    CONTACT_RIGHT,
    0,
    0,
    CONTACT_LEFT + CONTACT_TOP,
    0,
    CONTACT_LEFT,
    0,
    CONTACT_TOP,
    0,
    0,
    CONTACT_FULL,
]


"""
    Class that defines a bounding box for objects like Line, Triangle,
    Rectangle, Circle or Polygon.

    The aim of the bounding box is to check if two objects intersects
    by first testing if their bounding boxes overlap and then further
    precisely check if there is a collision of objects

    A bounding box is defined by two points : the tl=(top,left) and
    br=(bottom,right) points

"""


class BoundingBox(object):

    """
    Constructor that defines a bounding box by providing
    the coordinates of the (left,top) and (right,bottom)
    points where:
    left <= right
    top >= bottom

    Note that we always give x before y
    """

    def __init__(self, left, top, right, bottom):
        assert left <= right
        assert top >= bottom
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom
        # compute width and height of bounding box
        self.width = right - left
        self.height = top - bottom

    """
        Check if the bounding box bb, passed as a parameter, is contained
        by this bounding box

        return True if bb is contained in this box, False otherwise

    """

    def contains(self, bb):
        # print("contains self ", self.x1, self.y1, self.x2, self.y2)
        # print("contains bb   ", bb.x1, bb.y1, bb.x2, bb.y2)

        if bb.left < self.left or bb.right > self.right:
            # print("here 1")
            return False
        if bb.top > self.top or bb.bottom < self.bottom:
            # print("here 2")
            return False
        return True

    """
        Check if a point is inside a bouding box

        Return True if it is the case, False otherwise

    """

    def point_is_inside(self, x, y=None):
        if y == None:
            # then x should be a Point with x and y coordinates
            return ((self.left <= x.x) and (x.x <= self.right)) and (
                (self.bottom <= x.y) and (x.y <= self.top)
            )
        else:
            return ((self.left <= x) and (x <= self.right)) and (
                (self.bottom <= y) and (y <= self.top)
            )

    """
        Check if two bounding boxes are in contact and also where the contact
        occurr ans especially if the bounding box passed as a parameter is in
        contat with this bounding box.

        This function returns two values:
        - the first value is a boolean that indicates that there is a contact
        - the second is an integer that indicates where the contact occurred

        See the CONTACT_LEFT, CONTACT_RIGHT, ... constants

    """

    def contact(self, bb):
        if (bb.bottom > self.top) or (bb.top < self.bottom):
            return False, 0

        if (bb.right < self.left) or (bb.left > self.right):
            return False, 0

        ptl = self.is_point_inside(bb.left, bb.top)
        ptr = self.is_point_inside(bb.right, bb.top)
        pbl = self.is_point_inside(bb.left, bb.bottom)
        pbr = self.is_point_inside(bb.right, bb.bottom)

        # compute contact value as a formula
        where = int(ptl) + 2 * int(ptr) + 4 * int(pbl) + 8 * int(pbr)

        # print(str(ptl), str(ptr), str(pbl), str(pbr), where)

        return True, where_exactly[where]

    """
        Draw a bounding box in a pygame.Surface
    """

    def draw(self, window, color=(255, 255, 255), width=1):
        pg.draw.rect(
            window, color, pg.Rect(self.left, self.top, self.width, self.height), 1
        )

    """
        Return bounding box as string
    """

    def __str__(self):
        return (
            "("
            + str(self.left)
            + ","
            + str(self.top)
            + ","
            + str(self.right)
            + ","
            + str(self.bottom)
            + ",w="
            + str(self.width)
            + ",h="
            + str(self.height)
            + ")"
        )


"""
    Base class for objects that can be drawn on a Canvas (see canvas.py)
    Each object must redefine the following methods:
    - rotate to apply a rotation to the object
    - translate to apply a translation to the object
    - get_bounding_box to return the bounding box of the object
"""


class Form(object):
    def __init__(self):
        pass

    def rotate(self, angle):
        pass

    def translate(self, pt):
        pass

    def get_bounding_box(self):
        pass

    def point_is_inside(self, pt):
        return False


"""
    Class that represents a point with its integer coordonates
"""


class Point(Form):
    def __init__(self, x, y):
        super().__init__()
        self.x = x
        self.y = y

    """
        Apply a rotation of angle radians on the point
    """

    def rotate(self, angle):
        xr = int(math.cos(angle) * self.x + math.sin(angle) * self.y)
        yr = int(-math.sin(angle) * self.x + math.cos(angle) * self.y)
        self.x = xr
        self.y = yr

    """
        Apply a translation on the point
    """

    def translate(self, pt):
        self.x += pt.x
        self.y += pt.y

    """
        Return bounding box of the point
    """

    def get_bounding_box(self):
        return BoundingBox(self.x, self.y, self.x + 1, self.y - 1)

    def __str__(self):
        return "(" + str(self.x) + "," + str(self.y) + ")"


"""
    Class that defines a Line made of two points
"""


class Line(Form):
    def __init__(self, pt1, pt2):
        super().__init__()
        self.pt1 = pt1
        self.pt2 = pt2

    def rotate(self, angle):
        self.pt1.rotate(angle)
        self.pt2.rotate(angle)

    def translate(self, pt):
        self.pt1.x += pt.x
        self.pt2.x += pt.x
        self.pt1.y += pt.y
        self.pt2.y += pt.y

    def get_bounding_box(self):
        x1 = min(self.pt1.x, self.pt2.x)
        y1 = max(self.pt1.y, self.pt2.y)
        x2 = max(self.pt1.x, self.pt2.x)
        y2 = min(self.pt1.y, self.pt2.y)

        return BoundingBox(x1, y1, x2, y2)

    def __str__(self):
        return "Line " + str(self.pt1) + " to " + str(self.pt2)


"""
    Rectangle donné par ses coins supérieur gauche (pt1) et
    inférieur droit (pt2)
"""


class Rectangle(Form):
    def __init__(self, tl, br):
        super().__init__()
        assert tl.x <= br.x
        assert tl.y >= br.y
        self.tl = tl
        self.br = br
        self.tr = Point(br.x, tl.y)
        self.bl = Point(tl.x, br.y)

        """
        print("Rectangle def: ")
        print(str(self.tl) + " " + str(self.br) +
              " " + str(self.tr) + " " + str(self.bl))
        """

        self.x1 = tl.x
        self.y1 = tl.y
        self.x2 = br.x
        self.y2 = br.y
        self.width = br.x - tl.x
        self.height = tl.y - br.y

        # print("width  "+str(self.width))
        # print("height "+str(self.height))

    def rotate(self, angle):
        self.tl.rotate(angle)
        self.br.rotate(angle)
        self.tr = Point(self.br.x, self.tl.y)
        self.bl = Point(self.tl.x, self.br.y)
        self.x1 = self.tl.x
        self.y1 = self.tl.y
        self.x2 = self.br.x
        self.y2 = self.br.y

    def translate(self, pt):
        self.tl.x += pt.x
        self.br.x += pt.x
        self.tr.y += pt.y
        self.br.y += pt.y
        self.tr = Point(self.br.x, self.tl.y)
        self.bl = Point(self.tl.x, self.br.y)
        self.x1 = self.tl.x
        self.y1 = self.tl.y
        self.x2 = self.br.x
        self.y2 = self.br.y

    def get_bounding_box(self):
        x1 = min(min(self.tl.x, self.br.x), min(self.tr.x, self.bl.x))
        y1 = max(max(self.tl.y, self.br.y), max(self.tr.y, self.bl.y))
        x2 = max(max(self.tl.x, self.br.x), max(self.tr.x, self.bl.x))
        y2 = min(min(self.tl.y, self.br.y), min(self.tr.y, self.bl.y))

        return BoundingBox(x1, y1, x2, y2)

    def __str__(self):
        return "Rectangle " + str(self.tl) + " to " + str(self.br)

    """
        Code taken from : 
        https://stackoverflow.com/questions/217578/how-can-i-determine-whether-a-2d-point-is-within-a-polygon
    """

    def point_is_inside(self, pt):
        l = [self.tl, self.tr, self.br, self.bl]
        xmin = min(l, key=lambda point: point.x).x
        ymin = min(l, key=lambda point: point.y).y
        xmax = max(l, key=lambda point: point.x).x
        ymax = max(l, key=lambda point: point.y).y

        # print("xmin=" + str(xmin))
        # print("ymin=" + str(ymin))
        # print("xmax=" + str(xmax))
        # print("ymax=" + str(ymax))

        # if outside bouding box return False
        if pt.x < xmin or pt.x > xmax or pt.y < ymin or pt.y > ymax:
            return False

        # if one of the edges return True
        for p in l:
            if pt.x == p.x and pt.y == p.y:
                return True

        # https://wrf.ecse.rpi.edu/Research/Short_Notes/pnpoly.html
        inside = False
        i = 0
        j = len(l) - 1
        while i < len(l):
            if ((l[i].y > pt.y) != (l[j].y > pt.y)) and pt.x < (l[j].x - l[i].x) * (
                pt.y - l[i].y
            ) / (l[j].y - l[i].y) + l[i].y:
                inside = not inside

            i = i + 1
            j = i

        return inside


class Triangle(Form):
    def __init__(self, v1, v2, v3):
        super().__init__()
        self.v1 = v1
        self.v2 = v2
        self.v3 = v3

    def rotate(self, angle):
        self.v1.rotate(angle)
        self.v2.rotate(angle)
        self.v3.rotate(angle)

    def translate(self, pt):
        self.v1.x += self.pt.x
        self.v1.y += self.pt.y
        self.v2.x += self.pt.x
        self.v2.y += self.pt.y
        self.v3.x += self.pt.x
        self.v3.y += self.pt.y

    def get_bounding_box(self):
        x1 = min(min(self.v1.x, self.v2.x), self.v3.x)
        y1 = max(max(self.v1.y, self.v2.y), self.v3.y)
        x2 = max(max(self.v1.x, self.v2.x), self.v3.x)
        y2 = min(min(self.v1.y, self.v2.y), self.v3.y)

        return BoundingBox(x1, y1, x2, y2)

    def __str__(self):
        return "Triangle " + str(self.v1) + ", " + str(self.v2) + ", " + str(self.v3)

    def sign(self, p1, p2, p3):
        return (p1.x - p3.x) * (p2.y - p3.y) - (p2.x - p3.x) * (p1.y - p3.y)

    """
        Code taken from:
        https://stackoverflow.com/questions/2049582/how-to-determine-if-a-point-is-in-a-2d-triangle
    """

    def point_is_inside(self, pt):
        d1 = self.sign(pt, self.v1, self.v2)
        d2 = self.sign(pt, self.v2, self.v3)
        d3 = self.sign(pt, self.v3, self.v1)

        has_neg = (d1 < 0) or (d2 < 0) or (d3 < 0)
        has_pos = (d1 > 0) or (d2 > 0) or (d3 > 0)

        return not (has_neg and has_pos)


class Circle(Form):
    def __init__(self, center, radius):
        super().__init__()
        assert radius > 0
        self.center = center
        self.radius = radius

    def rotate(self, angle):
        self.pt1.rotate(angle)
        self.pt2.rotate(angle)

    def translate(self, pt):
        self.center.x += self.pt.x
        self.center.y += self.pt.y

    def get_bounding_box(self):
        x1 = self.center.x - self.radius
        y1 = self.center.y + self.radius
        x2 = self.center.x + self.radius
        y2 = self.center.y - self.radius

        return BoundingBox(x1, y1, x2, y2)

    def __str__(self):
        return "Circle " + str(self.center) + ", " + str(self.radius)

    def point_is_inside(self, pt):
        dx = pt.x - self.center.x
        dy = pt.y - self.center.y
        return (dx * dx + dy * dy) <= self.radius * self.radius


class Polygon(Form):
    def __init__(self, points):
        super().__init__()
        assert len(points) > 3
        self.points = points

    def rotate(self, angle):
        for p in self.points:
            p.rotate(angle)

    def translate(self, pt):
        for p in self.points:
            p.x += pt.x
            p.y += pt.y

    def get_bounding_box(self):
        x1 = self.points[0].x
        y1 = self.points[0].y
        x2 = self.points[0].x
        y2 = self.points[0].y

        for i in range(1, len(self.points)):
            x1 = min(x1, self.points[i].x)
            y1 = max(y1, self.points[i].y)
            x2 = max(x2, self.points[i].x)
            y2 = min(y2, self.points[i].y)

        return BoundingBox(x1, y1, x2, y2)

    def __str__(self):
        return "Polygon "

    def point_is_inside(self, pt):
        i = 0
        j = len(self.points) - 1
        c = False
        while i < len(self.points):
            c1 = (self.points[i].y > pt.y) != (self.points[j].y > pt.y)
            v1 = (self.points[j].x - self.points[i].x) * (pt.y - self.points[i].y)
            v2 = self.points[j].y - self.points[i].y + self.points[i].x
            c2 = pt.x < (v1 / v2)
            if c1 and c2:
                c = not c
            i = i + 1
            j = i
        return c


# ===================================================================
#
#
#
# ===================================================================


"""
    Class that defines a rotation of an angle given in radians.

    This class is used with the Canvas class (see canvas.py)
    in order to apply transformations to a Form.
"""


class Rotation(object):
    def __init__(self, angle):
        self.angle = angle

    def apply(self, object):
        object.rotate(self.angle)


"""
    Class that defines a translation.

    This class is used with the Canvas class (see canvas.py)
    in order to apply transformations to a Form.
"""


class Translation(object):
    def __init__(self, pt):
        self.pt = pt

    def apply(self, object):
        object.translate(self.pt)
